module SKUI
  # @since 1.0.0
  class Base

    require File.join( PATH, 'enum_system_color.rb' )
    require File.join( PATH, 'enum_system_font.rb' )
    require File.join( PATH, 'events.rb' )
    require File.join( PATH, 'font.rb' )
    require File.join( PATH, 'json.rb' )
    require File.join( PATH, 'properties.rb' )
    require File.join( PATH, 'typecheck.rb' )

    include Events
    extend Properties

    # ID string used by both Ruby and the WebDialog to keep each control in
    # sync with each other when passing properties and events.
    #
    # @return [String]
    # @since 1.0.0
    prop_reader( :ui_id ) # (i) :id would conflict with Object.id

    # @return [ControlManager, Nil]
    # @since 1.0.0
    prop( :parent, &TypeCheck::CONTAINER )

    # @return [Font, SystemFont]
    # @since 1.0.0
    prop( :font, &TypeCheck::FONT )

    # @return [Sketchup::Color, SystemColor]
    # @since 1.0.0
    prop( :foreground_color, &TypeCheck::COLOR )

    # @return [Sketchup::Color, SystemColor]
    # @since 1.0.0
    prop( :background_color, &TypeCheck::COLOR )

    # @return [JSON]
    # @since 1.0.0
    attr_accessor( :properties )

    # @return [Window, Nil]
    # @since 1.0.0
    attr_accessor( :window )
    
    #Color
    $color_fill_disabled = [220,220,220]
    $color_text_disabled = "darkgray"
    
    
    # @since 1.0.0
    def initialize
      super()
      # @properties contains all the data that must be shared with the webdialog
      # in order to sync everything on both ends.
      @properties = JSON.new
      @properties[ :ui_id ] = "UI_#{object_id()}"
      @properties[ :type ] = typename()
    end

    # @return [String]
    # @since 1.0.0
    def inspect
      "<#{self.class}:#{object_id_hex}>"
    end

    # @return [String]
    # @since 1.0.0
    def to_js
      ui_id.inspect
    end

    # Release all references to other objects. Setting them to nil. So that
    # the GC can collect them.
    #
    # @return [Nil]
    # @since 1.0.0
    def release
      release_events()
      @properties.clear
      @properties = nil
      @parent = nil
      nil
    end

    # @return [String]
    # @since 1.0.0
    def typename
      self.class.to_s.split( '::' ).last
    end

    # @return [Window|Nil]
    # @since 1.0.0
    def window
      control = self
      until control.parent.nil?
        control = control.parent
      end
      # Unless the root element is a Window then the control hasn't been added
      # to the window yet.
      ( control.is_a?( Window ) ) ? control : nil
    end

    private

    # @return [String]
    # @since 1.0.0
    def object_id_hex
      "0x%x" % ( object_id << 1 )
    end

    private

    # Call this method whenever a control property changes, spesifying which
    # properties changed. This is sent to the WebDialog for syncing.
    #
    # @param [Symbol] properties
    #
    # @return [Boolean]
    # @since 1.0.0
    def update_properties( *properties )
      if window && window.visible?
        # These properties must always be included unmodified.
        base_properties = {
          :type => typename()
        }
        # The given properties will be sent to the WebDialog where it updates
        # the UI to match the state of the Ruby objects.
        control_properties = JSON.new
        for property in properties
          control_properties[ property ] = @properties[ property ]
        end
        control_properties.merge!( base_properties )
        window.bridge.call( 'UI.update_properties', ui_id, control_properties )
        true
      else
        false
      end
    end

  end # class

  #Draw2d GL ========================================================
  
  class CuricDraw2d

    def initialize()
      @origin = Geom::Point3d.new(0, 0, 0)
      @vecZ = Geom::Vector3d.new(0, 0, 1)
    end
    
    def position_to(pts, position)
      point = Geom::Point3d.new(position[0], position[1], 0)
      tr = Geom::Transformation.new(point)
      pts.each{|p|
        p.transform! tr
      }
    end
    
    def rotate_to(pts, angle, origin=Geom::Point3d.new(0, 0, 0))
      angle = angle.degrees
      tr = Geom::Transformation.rotation(origin, Geom::Vector3d.new(0, 0, 1), angle)
      pts.each{|p|
        p.transform! tr
      }
    end
    
    def scale_to(pts, angle, origin=Geom::Point3d.new(0, 0, 0))
      angle = angle.degrees
      tr = Geom::Transformation.rotation(origin, Geom::Vector3d.new(0, 0, 1), angle)
      pts.each{|p|
        p.transform! tr
      }
    end
    
    def stroke_size(pts, size=-2)
      return offset_point(-size, pts)
    end
    
    def offset_point(dist,list)
        begin
            pi = Math::PI
            if (not ((dist.class==Fixnum || dist.class==Float || dist.class==Length) && dist!=0))
              return nil
            end
            verts=list
            pts = []
            0.upto(verts.length-1) do |a|
              vec1 = (verts[a]-verts[a-(verts.length-1)]).normalize
              vec2 = (verts[a]-verts[a-1]).normalize
              vec3 = (vec1+vec2).normalize
              if vec3.valid?
                ang = vec1.angle_between(vec2)/2
                ang = pi/2 if vec1.parallel?(vec2)
                vec3.length = dist/Math::sin(ang)
                t = Geom::Transformation.new(vec3)
                if pts.length > 0
                  vec4 = pts.last.vector_to(verts[a].transform(t))
                  if vec4.valid?
                    unless (vec2.parallel?(vec4))
                      t = Geom::Transformation.new(vec3.reverse)
                    end
                  end
                end
                
                pts.push(verts[a].transform(t))
              end
            end

            duplicates = []
            pts.each_index do |a|
              pts.each_index do |b|
                next if b==a
                duplicates<<b if pts[a]===pts[b]
              end
              break if a==pts.length-1
            end
            duplicates.reverse.each{|a| pts.delete(pts[a])}
            
            (pts.length > 2) ? (return pts) : (return nil)

        rescue
          puts "#{list} did not offset: #{list}"
          return nil
        end
      end
      
  end

  class Rectangle < CuricDraw2d

    attr_accessor( :height, :width , :stroke, :fill, 
        :color_stroke, :color_fill, :position, :rotate)
    
    def initialize()
      @height = 20
      @width = 50
      @stroke = 0
      @fill = false
      @color_stroke = "black"
      @color_fill = "red"
      @position = [0,0]
      @rotate = 0
    end
    
    def properties
      return {
        :height         => @height,
        :width          => @width,
        :stroke         => @stroke,
        :fill           => @fill,
        :color          => @color,
        :color_stroke   => @color_stroke,
        :color_fill     => @color_fill,
        :position       => @position
      }
    end
    
    def points
      pts=[Geom::Point3d.new(0,0,0),
        Geom::Point3d.new(@width,0,0),
        Geom::Point3d.new(@width,@height,0),
        Geom::Point3d.new(0,@height,0)]
      #position
      position_to(pts, @position)
      # rotate
      # rotate_to(pts, @rotate)
      return pts
    end
    
    
    def draw(view)
      view.line_width = 1
      rec = points
      
      if @fill
        view.drawing_color = @color_fill
        view.draw2d(GL_QUADS, rec)
      end
      
      if @stroke!=0
        # view.drawing_color = @color_stroke
        # view.draw2d(GL_LINE_LOOP, rec)
        view.drawing_color = @color_stroke
        rec2=stroke_size(rec, @stroke)
        view.draw2d(GL_LINE_LOOP, rec2)
        for i in 0..rec.length-1
          if i == rec.length-1
            quads = [rec[i],rec[0],rec2[0],rec2[i]]
            view.draw2d( GL_QUADS, quads )
          end
          quads = [rec[i],rec[i+1],rec2[i+1],rec2[i]]
          view.draw2d( GL_QUADS, quads )
        end
      end
    end
    
  end

  class Circle < CuricDraw2d

    attr_accessor( :diameter , :radius , :slide, :stroke, :fill, 
        :color_stroke, :color_fill, :position)
        
    attr_reader :radius	
    
    def initialize()
      @diameter = 20
      @radius = @diameter/2
      @slide = 36
      @stroke = 0
      @fill = false
      @color_stroke = "black"
      @color_fill = "red"
      @position = [0,0]
      @center=Geom::Point3d.new(0, 0, 0)
      @vecZ = Geom::Vector3d.new(0, 0, 1)
    end
    
    def properties
      return {
        :diameter       => @height,
        :radius         => @width,
        :stroke         => @stroke,
        :fill           => @fill,
        :color_stroke   => @color_stroke,
        :color_fill     => @color_fill,
        :position       => @position
      }
    end
    
    def points
      angle = (360/@slide).degrees
      transformation = Geom::Transformation.rotation(@center, @vecZ, angle)
      pts = [Geom::Point3d.new(@diameter, 0, 0)]
      for i in 0..@slide-2
        pts << (pts.last.transform( transformation ))
      end
      
      if pts.length!=0
        position_to(pts, @position)
        return pts 
      end
    end
    
    def draw(view)
      view.drawing_color = @color_stroke
      circle = points()
      if circle.length!=0
        if @fill
          view.drawing_color = @color_fill
          triangles = []
          for i in 1..circle.length-2
            triangles = [circle[0],circle[i],circle[i+1]]
            view.draw2d( GL_TRIANGLES, triangles )
          end
        end
        
        if @stroke!=0 && circle.length > 1
          view.drawing_color = @color_stroke
          circle2 = stroke_size(circle, @stroke)
          view.draw2d(GL_LINE_LOOP, circle2)
          for i in 0..circle.length-1
            if i == circle.length-1
              quads = [circle[i],circle[0],circle2[0],circle2[i]]
              view.draw2d( GL_QUADS, quads )
            end
            quads = [circle[i],circle[i+1],circle2[i+1],circle2[i]]
            view.draw2d( GL_QUADS, quads )
          end
        end
      end
    end 
    
  end

  class Arc < CuricDraw2d
      attr_accessor( :diameter , :radius , :type, :slide, :stroke, :fill, 
        :color_stroke, :color_fill, :position, :rotate)
        
      attr_reader :radius	
      
      def initialize()
        @diameter = 20
        @radius = @diameter/2
        @type = 0.5
        @slide = 36
        @stroke = 0
        @fill = false
        @color_stroke = "black"
        @color_fill = "red"
        @position = [0,0]
        @rotate = 0
        @center=Geom::Point3d.new(0, 0, 0)
        @vecZ = Geom::Vector3d.new(0, 0, 1)
      end
      
      def properties
        return {
          :diameter       => @height,
          :radius         => @width,
          :stroke         => @stroke,
          :fill           => @fill,
          :color_stroke   => @color_stroke,
          :color_fill     => @color_fill,
          :position       => @position
        }
      end
      
      def points
        angle = ((360*type)/@slide).degrees
        transformation = Geom::Transformation.rotation(@center, @vecZ, angle)
        pts = [Geom::Point3d.new(@diameter, 0, 0)]
        for i in 0..@slide-2
          pts << (pts.last.transform( transformation ))
        end
        
        if pts.length!=0
          rotate_to(pts, @rotate)
          position_to(pts, @position)
          return pts 
        end
      end
      
      def draw(view)
        view.drawing_color = @color_stroke
        arc = points()
        if arc.length!=0
          view.draw2d(GL_LINE_STRIP, arc)
          if @fill
            view.drawing_color = @color_fill
            triangles = []
            for i in 1..arc.length-2
              triangles = [arc[0],arc[i],arc[i+1]]
              view.draw2d( GL_TRIANGLES, triangles )
            end
          end
          
          if @stroke!=0 && arc.length > 1
            view.drawing_color = @color_stroke
            arc2 = stroke_size(arc, @stroke)
            view.draw2d(GL_LINE_STRIP, arc2)
            for i in 0..arc.length-2
              quads = [arc[i],arc[i+1],arc2[i+1],arc2[i]]
              view.draw2d( GL_QUADS, quads )
            end
            
            #fix start
            p01 = arc[0]
            p02 = arc[1]
            vecab = p01.vector_to( p02)
            vecz = Geom::Vector3d.new(0, 0, 1)
            vec = vecab.cross vecz
            p1 = arc[0].offset(vec, -@stroke)
            p2 = arc[1].offset(vec, -@stroke)
            view.draw2d(GL_QUADS, [p01,p1,p2,p02])
            quads = [arc[0],p1,p2,arc[1]]
            view.draw2d( GL_QUADS, quads )
            #fix end
            p01 = arc[-2]
            p02 = arc[-1]
            vecab = p01.vector_to( p02)
            vecz = Geom::Vector3d.new(0, 0, 1)
            vec = vecab.cross vecz
            p1 = arc[-2].offset(vec, -@stroke)
            p2 = arc[-1].offset(vec, -@stroke)
            view.draw2d(GL_QUADS, [p01,p1,p2,p02])
          end
        end
      end 
    end #Arc
    
  class Line < CuricDraw2d
    attr_accessor( :start, :end , :stroke, 
        :color_stroke)
    
    def initialize()
      @start = 20
      @end = 50
      @stroke = 0
      @color_stroke = "black"
    end
    def properties
      return {
        :start          => @start,
        :end            => @end,
        :stroke         => @stroke,
        :color_stroke   => @color_stroke
      }
    end
    
    def points
        pts=[Geom::Point3d.new(@start[0],@start[1],0),
        Geom::Point3d.new(@end[0],@end[1],0)]
    end
    
    def pend
        return points[0]
    end
    
    def pstart
        return points[1]
    end
    
    def draw(view)
      view.drawing_color = @color_stroke
      line = points()
      view.draw2d(GL_LINES, line)
      if @stroke!=0
        vecab = pstart.vector_to(pend)
        vecz = Geom::Vector3d.new(0, 0, 1)
        vec = vecab.cross vecz
        p1 = pstart.offset(vec, @stroke)
        p2 = pend.offset(vec, @stroke)
        view.draw2d(GL_QUADS, [pstart,p1,p2,pend])
      end
    end
  end#Line

  class Polyline < CuricDraw2d
    attr_accessor( :list , :stroke, 
        :color_stroke, :close, :fill, :color_fill)
    
    def initialize()
      @list = []
      @stroke = 0
      @color_stroke = "black"
      @close = false
      @fill = false
      @color_fill = "red"
    end
    
    def properties
      return {
        :list          => @list,
        :stroke         => @stroke,
        :color_stroke   => @color_stroke
      }
    end
    
    def points
      pts = []
      list.each{|xy|
        pts << Geom::Point3d.new(xy[0],xy[1],0)
      }
      return pts if pts != []
    end
    
    def pend
        return points[0]
    end
    
    def pstart
        return points[1]
    end
    
    def draw(view)
      view.drawing_color = @color_stroke
      polyline = points()
      if @stroke!=0
        for i in 0..polyline.length-2
          p01 = polyline[i]
          p02 = polyline[i+1]
          
          vecab = p01.vector_to( p02)
          vecz = Geom::Vector3d.new(0, 0, 1)
          vec = vecab.cross vecz
          p1 = polyline[i].offset(vec, @stroke)
          p2 = polyline[i+1].offset(vec, @stroke)
          view.draw2d(GL_QUADS, [p01,p1,p2,p02])
        
          quads = [polyline[i],p1,p2,polyline[i+1]]
          view.draw2d( GL_QUADS, quads )
        end
      end
      
      if @fill
        view.drawing_color = @color_fill
        triangles = []
        for i in 1..polyline.length-2
          triangles = [polyline[0],polyline[i],polyline[i+1]]
          view.draw2d( GL_TRIANGLES, triangles )
        end
      end
    end
    
  end#Polyline

  class Arrow < CuricDraw2d

    attr_accessor( :size, 
        :color_fill, :position, :rotate)
        
    attr_reader :radius	
    
    def initialize()
      @size = 10
      @color_fill = "red"
      @position = [0,0]
      @rotate = 0
    end
    
    def properties
      return {
        :color_fill     => @color_fill,
        :position       => @position,
        :rotate       => @rotate
      }
    end
    
    def points
      pts = [Geom::Point3d.new(0, 0, 0), 
          Geom::Point3d.new(@size, 0, 0), 
          Geom::Point3d.new(@size/2, @size/2, 0)]
      
      rotate_to(pts, @rotate)
      position_to(pts, @position)
      return pts 
    end
    
    def draw(view)
      view.drawing_color = @color_fill
      arrow = points()
      view.draw2d( GL_TRIANGLES, arrow )
    end 
    
  end

  class Soild < CuricDraw2d
      attr_accessor( :list , :stroke, 
          :color_stroke)
      
      def initialize()
        @list = []
        @stroke = 0
        @color_stroke = "black"
      end
      
      def properties
        return {
          :list          => @list,
          :stroke         => @stroke,
          :color_stroke   => @color_stroke
        }
      end
      
      def points
        pts = []
        list.each{|xy|
          pts << Geom::Point3d.new(xy[0],xy[1],0)
        }
        return pts if pts != []
      end
      
      def pend
          return points[0]
      end
      
      def pstart
          return points[1]
      end
      
      def draw(view)
        view.drawing_color = @color_stroke
        polyline = points()
        view.draw2d(GL_LINE_STRIP, polyline)
        if @stroke!=0
          for i in 0..polyline.length-2
            p01 = polyline[i]
            p02 = polyline[i+1]
            
            vecab = p01.vector_to( p02)
            vecz = Geom::Vector3d.new(0, 0, 1)
            vec = vecab.cross vecz
            p1 = polyline[i].offset(vec, @stroke)
            p2 = polyline[i+1].offset(vec, @stroke)
            view.draw2d(GL_QUADS, [p01,p1,p2,p02])
          
            quads = [polyline[i],p1,p2,polyline[i+1]]
            view.draw2d( GL_QUADS, quads )
          end
        end
      end
      
    end#Soild
  
  end # module
