module ShdsToolbar
  module DiscMenuEditor
    VERSION = '1.1.8'

    # This plugin's folder.
    PATH = File.dirname(__FILE__) unless defined?(self::PATH)

    # Requirements
    %w(translate.rb
       bridge.rb
       options.rb
       menus.rb
    ).each{ |file| require(File.join(PATH, file)) }

    EDITOR_HTML = File.join(PATH, 'html', "DiscMenuEditor.html") unless defined?(self::EDITOR_HTML)

    # Load translation strings and add plugin to UI.
    self::TRANSLATE = Translate.new('disc_menu_editor.strings') unless defined?(self::TRANSLATE)

    # Index of all commands.
    require(File.join(PATH, 'index.rb'))

    # Options.
    @options ||= Options.new(DiscMenuEditor, {
      :toolbars => {}, # id => { "name" => [String], "buttons" => [Array<Fixnum>] }
    })

    # Reference to the main dialog.
    @editor_dlg ||= nil

    public

    def self.get_menu_data_arr
      Menus.get_data
    end

    def self.import_config_file
      data_path = UI.openpanel(LH['选择圆盘菜单配置文件'], nil, 'Config File|*.dat;||')
      return if data_path.nil?
      json = File.read(data_path)
      data = JSON.parse(json)

      @options.options = {toolbars: {}}
      @options.update(data)
      @options.save
      Menus.clear
      self.load

      UI.messagebox(LH['导入成功'])
    end

    def self.export_config_file
      save_path = UI.savepanel(LH['保存圆盘菜单配置文件'], nil, "disc_menu_config.dat")
      return if save_path.nil?
      File.open(save_path, "w"){|f| f.write(@options.options.to_json)}
    end

    # This displays the main dialog with search field.
    def self.show_dialog
      # If the dialog exists, bring it to the front.
      Sketchup.active_model.select_tool(nil)
      if @editor_dlg && @editor_dlg.visible?
        @editor_dlg.bring_to_front
      else

        properties = {
            :dialog_title    => TRANSLATE["Toolbar Editor"],
            :preferences_key => "shds_toolbar_disc_menu_editor",
            :scrollable      => false,
            :resizable       => true,
            :width           => 600,
            :height          => 550,
            :left            => 600,
            :top             => 200,
        }
        if defined?(UI::HtmlDialog)
          properties[:style] = UI::HtmlDialog::STYLE_DIALOG
          @editor_dlg = UI::HtmlDialog.new(properties)
        else
          @editor_dlg = UI::WebDialog.new(properties)
        end
        @editor_dlg.set_file(EDITOR_HTML)

        # Add a Bridge to handle JavaScript-Ruby communication.
        @editor_dlg = Bridge.decorate(@editor_dlg)

        @editor_dlg.on("translate") { |action_context|
          # Translate.
          TRANSLATE.webdialog(@editor_dlg)
        }

        @editor_dlg.on("get_available_buttons") { |action_context|
          buttons = Index.instance.get_all{ |hash|
            hash[:icon].is_a?(String) && !hash[:icon].empty?
          }.sort_by{ |entry|
            entry[:name] || ""
          }.sort_by{ |entry|
            entry[:category] || ""
          }
          action_context.resolve(buttons)
        }

        @editor_dlg.on("get_toolbars") { |action_context|
          action_context.resolve(Menus.get_all)
        }

        # Save the toolbars and close the dialog.
        @editor_dlg.on("save_toolbars") { |action_context, hash|
          hash2 = {}
          hash.each{ |id, metadata|
            id = id.to_i
            Menus.save(id, metadata)
            hash2[id] = metadata
          }
          # Save the data.
          @options[:toolbars] = hash2
          @options.save
          @editor_dlg.close
        }

        # Delete a toolbar.
        @editor_dlg.on("delete_toolbar") { |action_context, id|
          Menus.delete(id)
          # Save the data.
          @options[:toolbars].delete(id)
          @options.save
        }

        # File dialog.
        @editor_dlg.on("openpanel") { |action_context, path|
          new_path = UI.openpanel(path)
          new_path = File.expand_path(new_path) if new_path
          action_context.resolve(new_path)
        }

        # Close the dialog.
        @editor_dlg.on("close") { |action_context|
          @editor_dlg.close
        }

        # Show the webdialog.
        @editor_dlg.show

      end
      return @editor_dlg
    end


    # Closes the main dialog if it is visible.
    def self.close
      if @editor_dlg && @editor_dlg.visible?
        @editor_dlg.close
      end
    end


    # Creates toolbar objects for commands remembered from @options and identified by IDs.
    def self.load
      # Load toolbars
      @options[:toolbars].each{ |id, metadata|
        next unless metadata.is_a?(Hash) && metadata["name"].is_a?(String) &&
          metadata["buttons"].is_a?(Array) && !metadata["buttons"].empty?
        Menus.load(id, metadata)
      }
    end

    unless file_loaded?(__FILE__)
      # Alternative 2: use timer that will execute after loading has finished.
      # Problem: SketchUp2013 doesn't remember UI::Toolbar.get_last_state for toolbars
      # that were created long after loading finished. This was fixed in 2013.
      UI.start_timer(0, false) {
        begin
          self.load()
        rescue Exception => error
          $stderr.write(error.message + $/)
          $stderr.write(error.backtrace.join($/) + $/)
        end
      }

      file_loaded(__FILE__)
    end # end unless
  end


end
