module ShdsToolbar


  module DiscMenuEditor


    module Menus
      Menu = Struct.new(:name, :commands, :command_int_hash)

      @@entries ||= {}
      @@menus ||= {}

      # Gets all entries of menus hashes.
      # @return [Array<Hash>]
      def self.get_all
        return @@entries.values.map{ |entry| entry.clone }
      end

      def self.clear
        @@entries = {}
        @@menus = {}
      end

      def self.get_data
        @@menus.map do |id, menu|
          command_int_hash = menu.command_int_hash
          sub_menu_data_arr = menu.commands.map do |c|
            [c.large_icon, c.menu_text, command_int_hash[c]]
          end
          [menu.name, sub_menu_data_arr]
        end
      end

      # Saves a new or modified menu.
      # @param [Fixnum] id
      # @param [Hash] metadata
      def self.save(id, metadata)
        if @@entries.include?(id) && @@menus.include?(id)
          self.update(id, metadata)
        else
          self.create(id, metadata)
        end
      end


      # We can use the same method when loading menus on startup
      def self.load(id, metadata)
        self.save(id, metadata)
      end


      # Creates a new menu.
      # @param [Fixnum] id
      # @param [Hash] metadata
      def self.create(id, metadata)
        # Create the entry.
        entry = {:id => id}
        update_entry(entry, metadata)
        menu = create_menu(metadata["name"], metadata["buttons"])
        @@menus[id] = menu
        @@entries[id] = entry
      end


      # Updates an existing menu.
      # @param [Fixnum] id
      # @param [Hash] metadata
      def self.update(id, metadata)
        # Update the entry.
        update_entry(@@entries[id], metadata)
        # Update the menu.
        update_menu(id, metadata["name"], metadata["buttons"])
      end


      # Deletes a menu.
      # @param [Fixnum] id
      def self.delete(id)
        @@entries.delete(id)
        @@menus.delete(id)
      end


      class << self


        # Updates the entry hash of a menu.
        # @param [Hash] entry
        # @param [Hash] metadata
        def update_entry(entry, metadata)
          entry.merge!(metadata)
        end
        private :update_entry


        # Creates a menu UI object.
        # @param [String] name
        # @param [Array<Fixnum>] button_ids
        # @return [UI::Toolbar]
        def create_menu(name, button_ids)
          commands, command_int_hash = [], {}
          menu = Menu.new(name, commands, command_int_hash)

          button_ids.each do |id|
            # Get the entry.
            entry = Index.instance.get_by_id(id)
            next if entry.nil?
            command = entry[:command] || UI::Command.new(entry[:name], &entry[:proc]) rescue next
            commands << command
            int = DiscMenuSubMenu.add_item(command)
            command_int_hash[command] = int
          end
          return menu
        end
        private :create_menu


        # Updates the menu UI object.
        # Since we can currently not remove commands from menus, we only add new ones.
        # @param [Fixnum] id
        # @param [String] name
        # @param [Array<Fixnum>] button_ids
        def update_menu(id, name, button_ids)
          menu = @@menus[id]
          commands = menu.commands
          command_int_hash = menu.command_int_hash
          button_ids.each do |id|
            entry = Index.instance.get_by_id(id)
            next if entry.nil?
            next if commands.find{|c| c == entry[:command] || c.is_a?(UI::Command) && c.menu_text == entry[:name]}
            command = entry[:command] || UI::Command.new(entry[:name], &entry[:proc]) rescue next
            commands << command
            int = DiscMenuSubMenu.add_item(command)
            command_int_hash[command] = int
          end
        end
        private :update_menu


      end # class << self


    end


  end


end
